// Common/StdOutStream.cpp

#include "StdAfx.h"

#include "StdOutStream.h"
#include "Common/IntToString.h"
#include "Common/StringConvert.h"
#include "Common/UTFConvert.h"

static const char kNewLineChar =  '\n';

CStdOutStream  g_StdOut;

CStdOutStream ::CStdOutStream ()
{
	m_lpOutBuf = (char*)::GlobalAlloc(GMEM_FIXED, 4096);
	m_lpCommandLine = NULL;
	m_lpArchiveName = NULL;
	m_bUnicode = 0;
	*m_lpTempSFXName = '\0';
	ReSet();
}

CStdOutStream ::~CStdOutStream ()
{
	::GlobalFree(m_lpOutBuf);
	if (m_lpCommandLine)
		free(m_lpCommandLine);
	if (m_lpArchiveName)
		free(m_lpArchiveName);
}

CStdOutStream & CStdOutStream::operator<<(CStdOutStream & (*aFunction)(CStdOutStream  &))
{
  (*aFunction)(*this);    
  return *this;
}

CStdOutStream & endl(CStdOutStream & outStream)
{
  return outStream << kNewLineChar;
}

CStdOutStream & CStdOutStream::operator<<(const char *string)
{
	AString strText = string;
	strText.Replace("\n", "\r\n");
	OutProgress(strText);
	return *this;
}

CStdOutStream & CStdOutStream::operator<<(const wchar_t *string)
{
	if (m_bUnicode)
	{
		AString strUtf8;
		::ConvertUnicodeToUTF8(string, strUtf8);
		*this << (const char *)strUtf8;
	}
	else
		*this << (const char *)UnicodeStringToMultiByte(string, CP_OEMCP);
	return *this;
}

CStdOutStream & CStdOutStream::operator<<(char c)
{
	char lpszText[3];
	if (c == '\n')
	{
		lpszText[0] = '\r';
		lpszText[1] = '\n';
		lpszText[2] = '\0';
	}
	else
	{
		lpszText[0] = c;
		lpszText[1] = '\0';
	}
	OutProgress(lpszText);
	return *this;
}

CStdOutStream & CStdOutStream::operator<<(int number)
{
	char textString[16];
	ConvertINT64ToString(number, textString);
	return operator<<(textString);
}

CStdOutStream & CStdOutStream::operator<<(UINT64 number)
{
  char textString[32];
  ConvertUINT64ToString(number, textString);
  return operator<<(textString);
}

// i󋵂o
void CStdOutStream::OutProgress(const char *lpszText)
{
	int nStrLen = strlen(lpszText);
	if (m_dwMaxBuf <= m_dwBufSize + nStrLen)
	{
		m_dwMaxBuf += 4096;
		m_lpOutBuf = (char*)::GlobalReAlloc(m_lpOutBuf, m_dwMaxBuf, GMEM_MOVEABLE);
	}
	memcpy(m_lpOutBuf + m_dwBufSize, lpszText, nStrLen + 1);
	m_dwBufSize += nStrLen;
}


// oϐ
void CStdOutStream::ReSet()
{
	m_pDlgProgress = NULL;
	m_hThread = NULL;
	m_lpOutBuf = (char*)::GlobalReAlloc(m_lpOutBuf, 4096, GMEM_MOVEABLE);
	m_dwBufSize = 0;
	m_dwMaxBuf = 4096;
	if (m_lpCommandLine)
	{
		free(m_lpCommandLine);
		m_lpCommandLine = NULL;
	}
	if (m_lpArchiveName)
	{
		free(m_lpArchiveName);
		m_lpArchiveName = NULL;
	}
}

// o͂Rs[
// lpCopyBufNULL̎͏o̓obt@̃|C^̂Ԃ
LPCSTR CStdOutStream::CopyBuf(LPSTR lpCopyBuf, DWORD dwBufSize)
{
	if (lpCopyBuf == NULL)
		return m_lpOutBuf;
	strncpy(lpCopyBuf, m_lpOutBuf, dwBufSize);
	lpCopyBuf[dwBufSize - 1] = '\0';
	return lpCopyBuf;
}

// ꂽR}hR}hCɂċL
void CStdOutStream::SetCommandLine(const CSplitCmdLine& scl)
{
	UString strCmdLine = L"7-zip32.dll";
	for (int i = 0; i < scl.m_argc; ++i)
	{
		strCmdLine += L" \"";
		strCmdLine += scl.m_argv[i];
		strCmdLine += L"\"";
	}
	if (scl.m_bSfx)
	{
		strCmdLine += L" \"-sfx";
		strCmdLine += GetTempSFXName();
		strCmdLine += L"\"";
	}
	if (m_lpCommandLine)
		free(m_lpCommandLine);
	m_lpCommandLine = _wcsdup(strCmdLine);
}

// LꂽR}hC擾
LPWSTR CStdOutStream::MyGetCommandLineW()
{
	return m_lpCommandLine;
}

// ꎞSFXt@C̎擾
LPCWSTR CStdOutStream::GetTempSFXName()
{
	if (*m_lpTempSFXName == '\0')
	{
		BOOL bIsNT;
		OSVERSIONINFO versionInfo;
		versionInfo.dwOSVersionInfoSize = sizeof(versionInfo);
		if (!::GetVersionEx(&versionInfo)) 
			bIsNT = FALSE;
		else
			bIsNT =  (versionInfo.dwPlatformId == VER_PLATFORM_WIN32_NT);

		if (bIsNT)
		{
			::GetTempPathW(MAX_PATH, m_lpTempSFXName);
			::GetTempFileNameW(m_lpTempSFXName, L"sfx", 0, m_lpTempSFXName);
		}
		else
		{
			char lpTempPath[MAX_PATH];
			::GetTempPath(MAX_PATH, lpTempPath);
			::GetTempFileName(lpTempPath, "sfx", 0, lpTempPath);
			wcscpy(m_lpTempSFXName, GetUnicodeString(lpTempPath));
		}
	}
	return m_lpTempSFXName;
}